import logging
from typing import Any

LOGGER_NAME = "causal_tracer"

logger = logging.getLogger(LOGGER_NAME)


def log_or_print(msg: Any, verbose: bool, level: int = logging.INFO) -> None:
    if verbose:
        print(msg)
    else:
        logger.log(level, msg)



from __future__ import annotations

from typing import Generator, Sequence, TypeVar

from tqdm import tqdm

T = TypeVar("T")


def find_all_substring_indices(
    string: str, substring: str, start: int = 0, end: int | None = None
) -> list[int]:
    """
    Find all indices of a substring in a string
    """
    indices = []
    while True:
        index = string.find(substring, start, end)
        if index == -1:
            break
        indices.append(index)
        start = index + len(substring)
    return indices


def batchify(
    data: Sequence[T], batch_size: int, show_progress: bool = False
) -> Generator[Sequence[T], None, None]:
    """Generate batches from data. If show_progress is True, display a progress bar."""

    for i in tqdm(
        range(0, len(data), batch_size),
        total=(len(data) // batch_size + (len(data) % batch_size != 0)),
        disable=not show_progress,
    ):
        yield data[i : i + batch_size]



from __future__ import annotations

from typing import Generic, TypeVar

T = TypeVar("T")


class PseudoFuture(Generic[T]):
    """
    A lightweight object that represents a future value.
    If the value is accessed before it's set an error is thrown.
    Just a way to have lighter-weight futures without needing asyncio.
    """

    _result: T | None = None

    def set_result(self, result: T) -> None:
        if self._result is not None:
            raise ValueError("Result is already set")
        self._result = result

    @property
    def result(self) -> T:
        if self._result is None:
            raise ValueError("Result is not set")
        return self._result



"""
Utilities for instrumenting a torch model.

Trace will hook one layer at a time.
TraceDict will hook multiple layers at once.
subsequence slices intervals from Sequential modules.
get_module, replace_module, get_parameter resolve dotted names.
set_requires_grad recursively sets requires_grad in module parameters.

Copied from https://github.com/kmeng01/rome/blob/main/util/nethook.py
"""

from __future__ import annotations

import contextlib
from typing import Any, Callable, Optional

import torch
from torch import nn

from causal_tracer.lib.torch_utils import get_module
from causal_tracer.lib.torch_utils import recursive_tensor_copy


class TraceLayer(contextlib.AbstractContextManager["TraceLayer"]):
    """
    To retain the output of the named layer during the computation of
    the given network:

        with Trace(net, 'layer.name') as ret:
            _ = net(inp)
            representation = ret.output

    A layer module can be passed directly without a layer name, and
    its output will be retained.  By default, a direct reference to
    the output object is returned, but options can control this:

        clone=True  - retains a copy of the output, which can be
            useful if you want to see the output before it might
            be modified by the network in-place later.
        detach=True - retains a detached reference or copy.  (By
            default the value would be left attached to the graph.)
        retain_grad=True - request gradient to be retained on the
            output.  After backward(), ret.output.grad is populated.

        retain_input=True - also retains the input.
        retain_output=False - can disable retaining the output.
        edit_output=fn - calls the function to modify the output
            of the layer before passing it the rest of the model.
            fn can optionally accept (output, layer) arguments
            for the original output and the layer name.
        stop=True - throws a StopForward exception after the layer
            is run, which allows running just a portion of a model.
    """

    input: Optional[torch.Tensor] = None
    output: Optional[torch.Tensor] = None

    def __init__(
        self,
        module: nn.Module,
        layer: str,
        retain_output: bool = True,
        retain_input: bool = False,
        clone: bool = False,
        detach: bool = False,
        retain_grad: bool = False,
        edit_output: Optional[Callable[[Any, str], Any]] = None,
        stop: bool = False,
    ):
        """
        Method to replace a forward method with a closure that
        intercepts the call, and tracks the hook so that it can be reverted.
        """
        retainer = self
        self.layer = layer
        layer_module = get_module(module, layer)

        def retain_hook(_m: Any, inputs: Any, output: Any) -> Any:
            if retain_input:
                retainer.input = recursive_tensor_copy(
                    inputs[0] if len(inputs) == 1 else inputs,
                    clone=clone,
                    detach=detach,
                    retain_grad=False,
                )  # retain_grad applies to output only.
            if edit_output:
                output = edit_output(output, self.layer)
            if retain_output:
                retainer.output = recursive_tensor_copy(
                    output, clone=clone, detach=detach, retain_grad=retain_grad
                )
                # When retain_grad is set, also insert a trivial
                # copy operation.  That allows in-place operations
                # to follow without error.
                if retain_grad and retainer.output is not None:
                    output = recursive_tensor_copy(
                        retainer.output, clone=True, detach=False
                    )
            if stop:
                raise StopForward()
            return output

        self.registered_hook = layer_module.register_forward_hook(retain_hook)
        self.stop = stop

    def __enter__(self) -> "TraceLayer":
        return self

    def __exit__(self, type: Any, _value: Any, _traceback: Any) -> bool | None:
        self.close()
        if self.stop and issubclass(type, StopForward):
            return True
        return None

    def close(self) -> None:
        self.registered_hook.remove()


class StopForward(Exception):
    """
    If the only output needed from running a network is the retained
    submodule then Trace(submodule, stop=True) will stop execution
    immediately after the retained submodule by raising the StopForward()
    exception.  When Trace is used as context manager, it catches that
    exception and can be used as follows:

    with Trace(net, layername, stop=True) as tr:
        net(inp) # Only runs the network up to layername
    print(tr.output)
    """

    pass



"""
Utilities for instrumenting a torch model.

Trace will hook one layer at a time.
TraceDict will hook multiple layers at once.
subsequence slices intervals from Sequential modules.
get_module, replace_module, get_parameter resolve dotted names.
set_requires_grad recursively sets requires_grad in module parameters.

Copied from https://github.com/kmeng01/rome/blob/main/util/nethook.py
"""

from __future__ import annotations

import contextlib
from collections import OrderedDict
from typing import Any, Callable, Iterable, Optional

from torch import nn

from causal_tracer.lib.TraceLayer import StopForward, TraceLayer


class TraceLayerDict(
    OrderedDict[str, TraceLayer],
    contextlib.AbstractContextManager["TraceLayerDict"],
):
    """
    To retain the output of multiple named layers during the computation
    of the given network:

        with TraceDict(net, ['layer1.name1', 'layer2.name2']) as ret:
            _ = net(inp)
            representation = ret['layer1.name1'].output

    If edit_output is provided, it should be a function that takes
    two arguments: output, and the layer name; and then it returns the
    modified output.

    Other arguments are the same as Trace.  If stop is True, then the
    execution of the network will be stopped after the last layer
    listed (even if it would not have been the last to be executed).
    """

    def __init__(
        self,
        module: nn.Module,
        layers: Optional[Iterable[str]] = None,
        retain_output: bool = True,
        retain_input: bool = False,
        clone: bool = False,
        detach: bool = False,
        retain_grad: bool = False,
        edit_output: Optional[Callable[[Any, str], Any]] = None,
        stop: bool = False,
    ):
        self.stop = stop

        def flag_last_unseen(it: Any) -> Any:
            try:
                it = iter(it)
                prev = next(it)
                seen = set([prev])
            except StopIteration:
                return
            for item in it:
                if item not in seen:
                    yield False, prev
                    seen.add(item)
                    prev = item
            yield True, prev

        for is_last, layer in flag_last_unseen(layers):
            self[layer] = TraceLayer(
                module=module,
                layer=layer,
                retain_output=retain_output,
                retain_input=retain_input,
                clone=clone,
                detach=detach,
                retain_grad=retain_grad,
                edit_output=edit_output,
                stop=stop and is_last,
            )

    def __enter__(self) -> "TraceLayerDict":
        return self

    def __exit__(self, type: Any, _value: Any, _traceback: Any) -> bool | None:
        self.close()
        if self.stop and issubclass(type, StopForward):
            return True
        return None

    def close(self) -> None:
        for _layer, trace in reversed(self.items()):
            trace.close()



from __future__ import annotations
from typing import Any, Optional, TypeVar

import torch
from torch import nn


def get_module(model: nn.Module, name: str) -> nn.Module:
    """
    Finds the named module within the given model.
    """
    for n, m in model.named_modules():
        if n == name:
            return m
    raise LookupError(name)


def set_requires_grad(requires_grad: bool, *models: nn.Module) -> None:
    """
    Sets requires_grad true or false for all parameters within the
    models passed.
    """
    for model in models:
        if isinstance(model, nn.Module):
            for param in model.parameters():
                param.requires_grad = requires_grad
        elif isinstance(model, (nn.Parameter, torch.Tensor)):
            model.requires_grad = requires_grad
        else:
            raise TypeError(f"unknown type {type(model)}")


def untuple_tensor(x: torch.Tensor | tuple[torch.Tensor, ...]) -> torch.Tensor:
    return x[0] if isinstance(x, tuple) else x


T = TypeVar("T", torch.Tensor, dict[Any, Any], list[Any], tuple[Any, ...])


def recursive_tensor_copy(
    x: T,
    clone: Optional[bool] = None,
    detach: Optional[bool] = None,
    retain_grad: Optional[bool] = None,
) -> T:
    """
    Copies a reference to a tensor, or an object that contains tensors,
    optionally detaching and cloning the tensor(s).  If retain_grad is
    true, the original tensors are marked to have grads retained.
    """
    if not clone and not detach and not retain_grad:
        return x
    if isinstance(x, torch.Tensor):
        if retain_grad:
            if not x.requires_grad:
                x.requires_grad = True
            x.retain_grad()
        elif detach:
            x = x.detach()
        if clone:
            x = x.clone()
        return x
    # Only dicts, lists, and tuples (and subclasses) can be copied.
    if isinstance(x, dict):
        return type(x)({k: recursive_tensor_copy(v) for k, v in x.items()})
    elif isinstance(x, (list, tuple)):
        return type(x)([recursive_tensor_copy(v) for v in x])
    else:
        assert False, f"Unknown type {type(x)} cannot be broken into tensors."



from __future__ import annotations

from collections import OrderedDict, defaultdict
from dataclasses import dataclass
from typing import Optional, Sequence
import torch
from torch import nn
from tokenizers import Tokenizer

from causal_tracer.causal_tracing.AsyncTracePatchProcessor import (
    AsyncTracePatchProcessor,
)
from causal_tracer.causal_tracing.LayerConfig import LayerConfig, get_layer_config
from causal_tracer.lib.PseudoFuture import PseudoFuture
from causal_tracer.causal_tracing.guess_subject import guess_subject
from causal_tracer.causal_tracing.pick_noise_level import pick_noise_level
from causal_tracer.lib.TraceLayerDict import TraceLayerDict
from causal_tracer.lib.layer_matching import (
    collect_matching_layers,
    get_layer_name,
)
from causal_tracer.lib.torch_utils import untuple_tensor
from causal_tracer.lib.constants import DEFAULT_DEVICE

from causal_tracer.lib.token_utils import (
    decode_tokens,
    find_token_range,
    make_inputs,
    predict_from_input,
)
from causal_tracer.lib.util import batchify
from .HiddenFlow import HiddenFlow, LayerKind


@dataclass
class BaseOutputs:
    answer_token_ids: torch.Tensor
    base_scores: torch.Tensor
    uncorrupted_layer_outputs: dict[int, OrderedDict[str, torch.Tensor]]
    attention_masks: list[torch.Tensor]


@dataclass
class HiddenFlowQuery:
    text: str
    subject: Optional[str] = None
    override_answer_id: Optional[int] = None


class CausalTracer:
    model: nn.Module
    tokenizer: Tokenizer
    layer_config: LayerConfig
    noise: float
    device: torch.device
    noise_calculation_subjects: Optional[list[str]]

    def __init__(
        self,
        model: nn.Module,
        tokenizer: Tokenizer,
        layer_config: Optional[LayerConfig] = None,
        noise: Optional[float] = None,
        noise_calculation_subjects: Optional[list[str]] = None,
        device: torch.device = DEFAULT_DEVICE,
    ) -> None:
        model.eval()
        self.model = model
        self.tokenizer = tokenizer
        if layer_config is None:
            layer_config = get_layer_config(model)
        self.layer_config = layer_config
        self.noise = noise or pick_noise_level(model, self.layer_config.embedding_layer)
        self.device = device
        self.noise_calculation_subjects = noise_calculation_subjects

    def count_layers(self) -> int:
        return len(
            collect_matching_layers(self.model, self.layer_config.hidden_layers_matcher)
        )

    def get_layer_name(self, layer_num: int, layer_kind: LayerKind) -> str:
        if layer_kind == "hidden":
            return get_layer_name(
                self.model, self.layer_config.hidden_layers_matcher, layer_num
            )
        if layer_kind == "mlp":
            return get_layer_name(
                self.model, self.layer_config.mlp_layers_matcher, layer_num
            )
        if layer_kind == "attention":
            return get_layer_name(
                self.model, self.layer_config.attention_layers_matcher, layer_num
            )
        raise ValueError(f"Unknown layer kind {layer_kind}")

    def get_layer_names_of_kind(self, layer_kind: LayerKind) -> list[str]:
        return [
            self.get_layer_name(layer_num, layer_kind)
            for layer_num in range(self.count_layers())
        ]

    def trace_important_window(
        self,
        prompt: str,
        uncorrupted_layer_outputs: dict[str, torch.Tensor],
        subject_range: tuple[int, int],
        answer_token_id: torch.Tensor,
        kind: LayerKind,
        trace_patch_processor: AsyncTracePatchProcessor,
        ntoks: int,
        window: int = 10,
        token_range: Optional[tuple[int, int]] = None,
        start_layer: int = 0,
        end_layer: Optional[int] = None,
    ) -> list[list[PseudoFuture[torch.Tensor]]]:
        """
        This appears to restore multiple layers at once, in a window around the layer of interest.
        This seems to only be used for the mlp or attention layers, never hidden layer.

        In the ROME colab, it says the following:

        'Because MLP and attention make small residual contributions, to observe a causal effect in those cases,
        we need to restore several layers of contributions at once, which is done by trace_important_window'
        """
        num_layers = self.count_layers()
        table: list[list[PseudoFuture[torch.Tensor]]] = []
        for tnum in range(*(token_range or (0, ntoks))):
            row = []
            for layer in range(start_layer, end_layer or num_layers):
                layerlist = [
                    (tnum, self.get_layer_name(layer_num, kind))
                    for layer_num in range(
                        max(0, layer - window // 2),
                        min(num_layers, layer - (-window // 2)),
                    )
                ]
                res_future = trace_patch_processor.trace_with_patch(
                    prompt,
                    states_to_patch=layerlist,
                    uncorrupted_layer_outputs=uncorrupted_layer_outputs,
                    answer_token_id=answer_token_id,
                    subject_range=subject_range,
                )
                row.append(res_future)
            table.append(row)
        return table

    def calculate_hidden_flow(
        self,
        prompt: str,
        subject: Optional[str] = None,
        kind: LayerKind = "hidden",
        answer_id_override: Optional[int] = None,
        samples: int = 10,
        window: int = 1,
        patch_seed: int = 1,
        batch_size: int = 32,
        start_layer: int = 0,
        end_layer: Optional[int] = None,
        patch_subject_tokens_only: bool = False,
    ) -> HiddenFlow:
        """
        Runs causal tracing over every token/layer combination in the network
        and returns a dictionary numerically summarizing the results.
        """
        return self.calculate_hidden_flows(
            [HiddenFlowQuery(prompt, subject, answer_id_override)],
            samples=samples,
            window=window,
            kind=kind,
            patch_seed=patch_seed,
            batch_size=batch_size,
            start_layer=start_layer,
            end_layer=end_layer,
            patch_subject_tokens_only=patch_subject_tokens_only,
        )[0]

    def calculate_base_outputs_batched(
        self,
        prompts: Sequence[str],
        kind: LayerKind,
        batch_size: int,
        show_progress: bool = False,
        move_to_cpu: bool = False,
        answer_id_overrides: list[tuple[int, int]] = [],
    ) -> BaseOutputs:
        # things will break if we try to process an empty batch, so just return empty results directly
        if len(prompts) == 0:
            return BaseOutputs(
                answer_token_ids=torch.empty(0),
                base_scores=torch.empty(0),
                uncorrupted_layer_outputs={},
                attention_masks=[],
            )
        answer_ids_outputs: list[torch.Tensor] = []
        base_scores_outputs: list[torch.Tensor] = []
        attention_masks: list[torch.Tensor] = []
        uncorrupted_layer_outputs: dict[
            int, OrderedDict[str, torch.Tensor]
        ] = defaultdict(OrderedDict)
        total_processed = 0
        for batch in batchify(prompts, batch_size, show_progress=show_progress):
            inputs = make_inputs(self.tokenizer, batch, self.device)
            with torch.no_grad(), TraceLayerDict(
                self.model,
                layers=self.get_layer_names_of_kind(kind),
                retain_output=True,
            ) as td:
                batch_answer_id_overrides = []
                for index, answer_id in answer_id_overrides:
                    if index >= total_processed and index < total_processed + len(
                        batch
                    ):
                        batch_answer_id_overrides.append(
                            (index - total_processed, answer_id)
                        )
                answer_token_ids, base_scores = predict_from_input(
                    self.model, inputs, batch_answer_id_overrides
                )
                raw_attn_masks = inputs["attention_mask"].detach().cpu()
                if move_to_cpu:
                    answer_token_ids = answer_token_ids.detach().cpu()
                    base_scores = base_scores.detach().cpu()
                answer_ids_outputs.append(answer_token_ids)
                base_scores_outputs.append(base_scores)
                for name, trace in td.items():
                    assert trace.output is not None  # keep mypy happy
                    raw_output = untuple_tensor(trace.output).detach()
                    if move_to_cpu:
                        raw_output = raw_output.cpu()
                    for input_num in range(len(batch)):
                        attn_indices = raw_attn_masks[input_num].nonzero(as_tuple=True)[
                            0
                        ]
                        uncorrupted_layer = raw_output[input_num, attn_indices]

                        uncorrupted_layer_outputs[total_processed + input_num][
                            name
                        ] = uncorrupted_layer
            total_processed += len(batch)
            for attention_mask in inputs["attention_mask"].detach().cpu():
                attention_masks.append(attention_mask)
        return BaseOutputs(
            answer_token_ids=torch.cat(answer_ids_outputs),
            base_scores=torch.cat(base_scores_outputs),
            uncorrupted_layer_outputs=uncorrupted_layer_outputs,
            attention_masks=attention_masks,
        )

    def calculate_hidden_flows(
        self,
        queries: list[str | HiddenFlowQuery],
        samples: int = 10,
        window: int = 1,
        kind: LayerKind = "hidden",
        patch_seed: int = 1,
        batch_size: int = 32,
        start_layer: int = 0,
        end_layer: Optional[int] = None,
        patch_subject_tokens_only: bool = False,
    ) -> list[HiddenFlow]:
        prompts: list[str] = []
        subjects: list[str] = []
        answer_id_overrides: list[tuple[int, int]] = []
        for i, query in enumerate(queries):
            if isinstance(query, str):
                prompts.append(query)
                subjects.append(guess_subject(query))
            else:
                prompts.append(query.text)
                subjects.append(query.subject or guess_subject(query.text))
                if query.override_answer_id is not None:
                    answer_id_overrides.append((i, query.override_answer_id))
        trace_patch_processor = AsyncTracePatchProcessor(
            self.model,
            self.tokenizer,
            self.layer_config.embedding_layer,
            self.noise,
            samples_per_patch=samples,
            batch_size=batch_size,
            device=self.device,
            random_seed=patch_seed,
        )
        base_outputs = self.calculate_base_outputs_batched(
            prompts, kind, batch_size, answer_id_overrides=answer_id_overrides
        )
        uncorrupted_layer_outputs = base_outputs.uncorrupted_layer_outputs

        answers = decode_tokens(self.tokenizer, base_outputs.answer_token_ids)
        inputs = make_inputs(self.tokenizer, prompts, self.device)
        subject_ranges = [
            find_token_range(self.tokenizer, input_ids, subject)
            for input_ids, subject in zip(inputs["input_ids"], subjects)
        ]
        ntoks = inputs["attention_mask"].sum(dim=1).tolist()
        hidden_flows = []
        for index, prompt in enumerate(prompts):
            subject_range = subject_ranges[index]
            token_range = subject_range if patch_subject_tokens_only else None
            answer_token_id = base_outputs.answer_token_ids[index]
            low_score_future = trace_patch_processor.trace_with_patch(
                prompt,
                states_to_patch=[],
                uncorrupted_layer_outputs=uncorrupted_layer_outputs[index],
                answer_token_id=answer_token_id,
                subject_range=subject_range,
            )
            differences_futures = self.trace_important_window(
                prompt,
                uncorrupted_layer_outputs[index],
                subject_range,
                answer_token_id,
                window=window,
                kind=kind,
                trace_patch_processor=trace_patch_processor,
                ntoks=ntoks[index],
                start_layer=start_layer,
                end_layer=end_layer,
                token_range=token_range,
            )
            # need to call this to resolve all the patch trace outputs
            trace_patch_processor.process()
            low_score = low_score_future.result.item()
            differences = self._calculate_differences(
                differences_futures,
                low_score,
                start_layer=start_layer,
                end_layer=end_layer,
                ntoks=ntoks[index],
                token_range=token_range,
            )
            hidden_flows.append(
                HiddenFlow(
                    scores=differences,
                    low_score=low_score,
                    high_score=base_outputs.base_scores[index].item(),
                    input_ids=inputs["input_ids"][index],
                    input_tokens=decode_tokens(
                        self.tokenizer, inputs["input_ids"][index]
                    ),
                    subject_range=subject_range,
                    answer=answers[index],
                    kind=kind,
                    layer_outputs=detach_layer_outputs(
                        uncorrupted_layer_outputs[index]
                    ),
                )
            )
        return hidden_flows

    def _calculate_differences(
        self,
        differences_futures: list[list[PseudoFuture[torch.Tensor]]],
        low_score: float,
        start_layer: int,
        end_layer: Optional[int],
        ntoks: int,
        token_range: Optional[tuple[int, int]],
    ) -> torch.Tensor:
        resolved_differences: list[torch.Tensor] = []
        for row in differences_futures:
            resolved_differences.append(torch.stack([item.result for item in row]))
        differences = torch.stack(resolved_differences).detach().cpu()

        # fill in any skipped layers with the low score
        if start_layer > 0:
            prepend_differences = torch.full(
                (differences.shape[0], start_layer), low_score
            )
            differences = torch.cat((prepend_differences, differences), dim=1)
        if end_layer is not None and end_layer < self.count_layers():
            append_differences = torch.full(
                (differences.shape[0], self.count_layers() - end_layer), low_score
            )
            differences = torch.cat((differences, append_differences), dim=1)

        if token_range is not None and token_range[0] > 0:
            prepend_differences = torch.full(
                (token_range[0], differences.shape[1]), low_score
            )
            differences = torch.cat((prepend_differences, differences), dim=0)
        if token_range is not None and token_range[1] < ntoks:
            append_differences = torch.full(
                (ntoks - token_range[1], differences.shape[1]), low_score
            )
            differences = torch.cat((differences, append_differences), dim=0)
        return differences


def detach_layer_outputs(
    layer_outputs: OrderedDict[str, torch.Tensor]
) -> OrderedDict[str, torch.Tensor]:
    return OrderedDict([(k, v.detach().cpu()) for k, v in layer_outputs.items()])



from causal_tracer.causal_tracing.CausalTracer import CausalTracer
from causal_tracer.causal_tracing.plot_hidden_flow_heatmap import (
    plot_hidden_flow_heatmap,
)
from causal_tracer.causal_tracing.LayerConfig import (
    LayerConfig,
    GPT2_LAYER_CONFIG,
    LLAMA_LAYER_CONFIG,
)

__all__ = [
    "CausalTracer",
    "LayerConfig",
    "GPT2_LAYER_CONFIG",
    "LLAMA_LAYER_CONFIG",
    "plot_hidden_flow_heatmap",
]



from typing import Optional
import os

from matplotlib import pyplot as plt

from .HiddenFlow import HiddenFlow, LayerKind


def plot_hidden_flow_heatmap(
    result: HiddenFlow,
    savepdf: Optional[str] = None,
    title: Optional[str] = None,
    xlabel: Optional[str] = None,
    modelname: Optional[str] = None,
    show: bool = True,
    font_family: str = "",
    color: Optional[str] = None,
) -> plt.Figure:
    differences = result.scores
    low_score = result.low_score
    answer = result.answer
    kind = result.kind
    labels = list(result.input_tokens)
    for i in range(*result.subject_range):
        labels[i] = labels[i] + "*"

    color_map: dict[LayerKind, str] = {
        "hidden": "Purples",
        "mlp": "Greens",
        "attn": "Reds",
    }

    with plt.rc_context(rc={"font.family": font_family}):
        fig, ax = plt.subplots(figsize=(3.5, 2), dpi=200)
        h = ax.pcolor(
            differences,
            cmap=color or color_map[kind],
            vmin=low_score,
        )
        ax.invert_yaxis()
        ax.set_yticks([0.5 + i for i in range(len(differences))])
        ax.set_xticks([0.5 + i for i in range(0, differences.shape[1] - 6, 5)])
        ax.set_xticklabels(list(range(0, differences.shape[1] - 6, 5)))
        ax.set_yticklabels(labels)
        if not modelname:
            modelname = "GPT"
        if kind == "hidden":
            ax.set_title("Impact of restoring state after corrupted input")
            ax.set_xlabel(f"single restored layer within {modelname}")
        else:
            kindname = kind
            ax.set_title(f"Impact of restoring {kindname} after corrupted input")
            ax.set_xlabel(f"center of interval of restored {kindname} layers")
        cb = plt.colorbar(h)
        if title is not None:
            ax.set_title(title)
        if xlabel is not None:
            ax.set_xlabel(xlabel)
        elif answer is not None:
            # The following should be cb.ax.set_xlabel, but this is broken in matplotlib 3.5.1.
            cb.ax.set_title(f"p({str(answer).strip()})", y=-0.16, fontsize=10)
        if savepdf:
            os.makedirs(os.path.dirname(savepdf), exist_ok=True)
            plt.savefig(savepdf, bbox_inches="tight")
            plt.close()
        if show:
            plt.show()
        return fig



from collections import defaultdict
from dataclasses import dataclass
from tokenizers import Tokenizer
import torch
import numpy as np
import numpy.typing as npt
from torch import nn

from causal_tracer.lib.PseudoFuture import PseudoFuture
from causal_tracer.lib.token_utils import make_inputs
from causal_tracer.lib.TraceLayerDict import TraceLayerDict
from causal_tracer.lib.torch_utils import untuple_tensor
from causal_tracer.lib.constants import DEFAULT_DEVICE


@dataclass
class ResultAccumulator:
    future: PseudoFuture[torch.Tensor]
    prompt: str
    result_parts: list[torch.Tensor]
    total_samples: int


@dataclass
class TracePatchSample:
    result_id: int
    prompt: str
    sample_num: int
    uncorrupted_layer_outputs: dict[str, torch.Tensor]
    states_to_patch: list[tuple[int, str]]
    subject_range: tuple[int, int]
    random_seed: int
    answer_token_id: torch.Tensor


class AsyncTracePatchProcessor:
    """
    Handles processing of patch traces in bulk
    """

    batch_size: int
    samples_per_patch: int
    embed_layername: str
    noise: float
    model: nn.Module
    tokenizer: Tokenizer
    results_tracker: dict[int, ResultAccumulator]
    work_queue: list[TracePatchSample]
    random_seed: int
    device: torch.device
    result_id_counter: int
    _noise_cache: dict[tuple[int, int, int], npt.NDArray[np.float64]]

    def __init__(
        self,
        model: nn.Module,
        tokenizer: Tokenizer,
        embed_layername: str,
        noise: float,
        samples_per_patch: int = 10,
        batch_size: int = 32,
        device: torch.device = DEFAULT_DEVICE,
        random_seed: int = 1,
    ) -> None:
        self.model = model
        self.batch_size = batch_size
        self.samples_per_patch = samples_per_patch
        self.tokenizer = tokenizer
        self.noise = noise
        self.device = device
        self.random_seed = random_seed
        self.embed_layername = embed_layername
        self.work_queue = []
        self.results_tracker = {}
        self._noise_cache = {}
        self.result_id_counter = 0

    def trace_with_patch(
        self,
        prompt: str,
        states_to_patch: list[tuple[int, str]],
        uncorrupted_layer_outputs: dict[str, torch.Tensor],
        answer_token_id: torch.Tensor,
        subject_range: tuple[int, int],
        random_seed: int = 1,
    ) -> PseudoFuture[torch.Tensor]:
        """
        Runs a single causal trace.  Given a model and a batch input, runs the batch in inference, corrupting
        a the set of runs [1...samples_per_patch] while also restoring a set of hidden states to
        the values from the uncorrupted run.

        The argument subject_range specifies an
        be corrupted by adding Gaussian noise to the embedding for the batch
        inputs other than the first element in the batch.  Alternately,
        subsequent runs could be corrupted by simply providing different
        input tokens via the passed input batch.

        Then when running, a specified set of hidden states will be uncorrupted
        by restoring their values to the same vector that they had in the
        zeroth uncorrupted run.  This set of hidden states is listed in
        states_to_patch, by listing [(token_index, layername), ...] pairs.
        To trace the effect of just a single state, this can be just a single
        token/layer pair.  To trace the effect of restoring a set of states,
        any number of token indices and layers can be listed.
        """
        result_id = self.result_id_counter
        self.result_id_counter += 1
        for sample_num in range(self.samples_per_patch):
            self.work_queue.append(
                TracePatchSample(
                    result_id=result_id,
                    sample_num=sample_num,
                    prompt=prompt,
                    uncorrupted_layer_outputs=uncorrupted_layer_outputs,
                    states_to_patch=states_to_patch,
                    subject_range=subject_range,
                    random_seed=random_seed,
                    answer_token_id=answer_token_id,
                )
            )
        result: PseudoFuture[torch.Tensor] = PseudoFuture()
        self.results_tracker[result_id] = ResultAccumulator(
            future=result,
            prompt=prompt,
            result_parts=[],
            total_samples=self.samples_per_patch,
        )
        self._process_if_full_batch()
        return result

    def process(self) -> None:
        """
        Runs the queued traces in batches
        """
        while len(self.work_queue) > 0:
            batch = self.work_queue[: self.batch_size]
            self.work_queue = self.work_queue[self.batch_size :]
            self._process_batch(batch)

    def _process_if_full_batch(self) -> None:
        """
        Runs a batch of traces if a full batch is available
        """
        while len(self.work_queue) >= self.batch_size:
            batch = self.work_queue[: self.batch_size]
            self.work_queue = self.work_queue[self.batch_size :]
            self._process_batch(batch)

    def _assert_non_mixed_batch(self, batch: list[TracePatchSample]) -> None:
        """
        Ensures that a batch of traces is not mixed between different prompts.
        This may be supported in the future, but for now it's best to assert it's not possible.
        We need to handle resolving different length outputs / predictions and attention masks
        before that can be supported.
        """
        assert (
            len({sample.prompt for sample in batch}) == 1
        ), "Batch contains mixed prompts. This is not currently supported. call process() before adding different prompts"

    def _process_batch(self, batch: list[TracePatchSample]) -> None:
        """
        Runs a single batch of traces
        """
        self._assert_non_mixed_batch(batch)
        prompts = [sample.prompt for sample in batch]
        all_layers_in_batch: set[str] = set()
        inputs = make_inputs(self.tokenizer, prompts, self.device)
        patch_spec_per_batch_index = []
        for sample in batch:
            patch_spec = defaultdict(list)
            for token, layer in sample.states_to_patch:
                all_layers_in_batch.add(layer)
                patch_spec[layer].append(token)
            patch_spec_per_batch_index.append(patch_spec)

        # Define the model-patching rule.
        def patch_rep(x: torch.Tensor, layer: str) -> torch.Tensor:
            if layer == self.embed_layername:
                # Corrupt a range of token embeddings on batch items x
                for batch_index, sample in enumerate(batch):
                    subj_start, subj_end = sample.subject_range
                    subj_len = subj_end - subj_start
                    consistent_noise = self._get_consistent_noise(
                        sample.sample_num, subj_len, x.shape[2]
                    )
                    noise = self.noise * torch.from_numpy(consistent_noise).to(x.device)
                    x[batch_index, subj_start:subj_end] += noise
                return x
            if layer not in all_layers_in_batch:
                return x
            # If this layer is in the patch_spec, restore the uncorrupted hidden state
            # for selected tokens.
            h = untuple_tensor(x)
            for batch_index, sample in enumerate(batch):
                patch_spec = patch_spec_per_batch_index[batch_index]
                for token in patch_spec[layer]:
                    h[batch_index, token] = sample.uncorrupted_layer_outputs[layer][
                        token
                    ]
            return x

        # With the patching rules defined, run the patched model in inference.
        with torch.no_grad(), TraceLayerDict(
            self.model,
            [self.embed_layername] + list(all_layers_in_batch),
            edit_output=patch_rep,
        ):
            outputs_exp = self.model(**inputs)

        for batch_index, sample in enumerate(batch):
            result_tracker = self.results_tracker[sample.result_id]
            result_tracker.result_parts.append(outputs_exp.logits[batch_index])
            # if we have all the parts of the result, compute the final result and set the future
            if len(result_tracker.result_parts) == result_tracker.total_samples:
                result_logits = torch.stack(result_tracker.result_parts)
                probs = torch.softmax(result_logits[:, -1, :], dim=1).mean(dim=0)[
                    sample.answer_token_id
                ]
                result_tracker.future.set_result(probs)
                self.results_tracker.pop(sample.result_id)

    def _get_consistent_noise(
        self, sample_num: int, subj_len: int, layer_size: int
    ) -> npt.NDArray[np.float64]:
        """
        Returns a numpy array of noise for a given sample number and shape, which is consistent for each combination of inputs
        """
        cache_key = (sample_num, subj_len, layer_size)
        if cache_key not in self._noise_cache:
            prng = np.random.RandomState(self.random_seed)
            self._noise_cache[cache_key] = prng.randn(
                self.samples_per_patch, subj_len, layer_size
            )[sample_num]
        return self._noise_cache[cache_key]



from __future__ import annotations
from typing import Callable, Union

from torch import nn

LayerMatcher = Union[str, Callable[[nn.Module, int], str]]


def collect_matching_layers(model: nn.Module, layer_matcher: LayerMatcher) -> list[str]:
    """
    Find all layers in the model that match the layer_matcher, in order by layer_num.
    layer_matcher can be a string formatted like "transformer.h.{num}.mlp" or a callable
    If layer_matcher is a callable, it should take in a model and layer_num and return
    a string representing the layer name corresponding to that layer number.
    If layer_matcher is a string, it's considered a template and MUST contain a "{num}" portion
    """
    matcher_callable = _layer_matcher_to_callable(layer_matcher)
    all_layer_names = dict(model.named_modules()).keys()
    matching_layers = []
    for layer_num, layer in enumerate(model.modules()):
        layer_name = matcher_callable(model, layer_num)
        if layer_name in all_layer_names:
            matching_layers.append(layer_name)
        else:
            break
    return matching_layers


def get_layer_name(
    model: nn.Module, layer_matcher: LayerMatcher, layer_num: int
) -> str:
    matcher_callable = _layer_matcher_to_callable(layer_matcher)
    return matcher_callable(model, layer_num)


def get_layer_by_name(model: nn.Module, layer_name: str) -> nn.Module:
    return dict(model.named_modules())[layer_name]


def _layer_matcher_to_callable(
    layer_matcher: LayerMatcher,
) -> Callable[[nn.Module, int], str]:
    if isinstance(layer_matcher, str):
        if "{num}" not in layer_matcher:
            raise ValueError(
                "layer_matcher must be a callable or a string containing {num}"
            )
        matcher_callable = lambda _model, layer_num: layer_matcher.format(num=layer_num)
        # for some reason mypy doesn't like directly returning the lambda without assigning to a var first
        return matcher_callable
    return layer_matcher



from dataclasses import dataclass

from torch import nn

from causal_tracer.lib.layer_matching import LayerMatcher


@dataclass
class LayerConfig:
    hidden_layers_matcher: LayerMatcher
    attention_layers_matcher: LayerMatcher
    mlp_layers_matcher: LayerMatcher
    embedding_layer: str


GPT2_LAYER_CONFIG = LayerConfig(
    hidden_layers_matcher="transformer.h.{num}",
    attention_layers_matcher="transformer.h.{num}.attn",
    mlp_layers_matcher="transformer.h.{num}.mlp",
    embedding_layer="transformer.wte",
)

LLAMA_LAYER_CONFIG = LayerConfig(
    hidden_layers_matcher="model.layers.{num}",
    attention_layers_matcher="model.layers.{num}.self_attn",
    mlp_layers_matcher="model.layers.{num}.mlp",
    embedding_layer="model.embed_tokens",
)


def get_layer_config(model: nn.Module) -> LayerConfig:
    if hasattr(model, "config"):
        config = model.config
        model_type: str = config.model_type
        if model_type == "gpt2":
            return GPT2_LAYER_CONFIG
        elif model_type == "llama":
            return LLAMA_LAYER_CONFIG
    raise ValueError(
        "Unknown model type. Please provide a LayerConfig to identify relevant model layers for tracing."
    )



from typing import cast
import torch
from torch import nn
from causal_tracer.lib.torch_utils import get_module


def pick_noise_level(
    model: nn.Module,
    embedding_layer: str = "transformer.wte",
    std_multiplier: float = 3,
) -> float:
    """
    Pick a noise level to corrupt the input text with, such that the
    noise is a multiplier of the stdev of the token embeddings.
    """
    with torch.no_grad():
        embedding_weights = cast(
            torch.Tensor, get_module(model, embedding_layer).weight
        )
        noise_level_std = embedding_weights.std().item()
    return std_multiplier * noise_level_std



import re


def guess_subject(prompt: str) -> str:
    match = re.search(r"(?!Wh(o|at|ere|en|ich|y) )([A-Z]\S*)(\s[A-Z][a-z']*)*", prompt)
    if not match:
        raise ValueError(f"Could not guess subject in prompt: {prompt}")
    return match[0].strip()



# from https://github.com/kmeng01/rome/blob/main/experiments/causal_trace.py

from __future__ import annotations
from typing import Iterable, Sequence

from tokenizers import Tokenizer
from transformers import BatchEncoding
import torch
from torch import nn

from causal_tracer.lib.constants import DEFAULT_DEVICE
from causal_tracer.lib.util import find_all_substring_indices


# Utilities for dealing with tokens
def make_inputs(
    tokenizer: Tokenizer,
    prompts: Sequence[str],
    device: torch.device = DEFAULT_DEVICE,
    add_pad_token: bool = True,
) -> dict[str, torch.Tensor]:
    ensure_tokenizer_has_pad_token(tokenizer, add_pad_token=add_pad_token)
    return tokenizer(prompts, padding=True, return_tensors="pt").to(device)


def ensure_tokenizer_has_pad_token(
    tokenizer: Tokenizer, add_pad_token: bool = True
) -> None:
    # from https://github.com/huggingface/transformers/issues/12594#issuecomment-877358955
    if not tokenizer.pad_token:
        if add_pad_token:
            tokenizer.pad_token = tokenizer.eos_token
        else:
            raise ValueError("Tokenizer must have a pad token")


def decode_tokens(
    tokenizer: Tokenizer, token_array: list[int] | torch.Tensor | list[torch.Tensor]
) -> list[str]:
    return [tokenizer.decode([t]) for t in token_array]


def get_answer_token_ids(
    tokenizer: Tokenizer,
    answer: str,
    ensure_space_prefix: bool = True,
    strip_start_token: bool = True,
    strip_blank_start_token: bool = True,
) -> list[int]:
    """
    Helper to find the token ids for the given answer as if it were a continuation of the prompt.
    """
    processed_answer = answer
    if ensure_space_prefix and not processed_answer.startswith(" "):
        processed_answer = " " + processed_answer
    tokens = tokenizer.encode(processed_answer)
    if strip_start_token and tokens[0] == tokenizer.bos_token_id:
        tokens = tokens[1:]
    # llama only includes an explicit space token at the start of the string if it's the first token
    if strip_blank_start_token and tokenizer.decode([tokens[0]]) == "":
        tokens = tokens[1:]
    return tokens


def tokenize_with_suffix_tokens(
    tokenizer: Tokenizer,
    prompts: Iterable[str],
    suffix_tokens: Iterable[Iterable[int]],
) -> BatchEncoding:
    """
    Helper to tokenize a list of prompts with a list of suffix tokens appended.
    This attempts to recreate similar functionality to tokenizer.__call__(),
    but with the ability to append suffix tokens while still maintaining the correct
    attention mask. This is much less efficient than using the built-in tokenizer
    """
    ensure_tokenizer_has_pad_token(tokenizer)
    prompts_tokens = tokenizer(prompts)["input_ids"]
    combined_tokens = [
        [*prompt_tokens, *suffix]
        for prompt_tokens, suffix in zip(prompts_tokens, suffix_tokens)
    ]
    return tokenizer.pad(
        {"input_ids": combined_tokens},
        padding=True,
        return_tensors="pt",
    )


def find_token_range(
    tokenizer: Tokenizer,
    token_array: list[int] | torch.Tensor,
    substring: str,
    find_last_match: bool = True,
) -> tuple[int, int]:
    # sometimes the tokenizer messes with non-alphanumeric characters
    # so make sure the substring goes through an encoding/decoding cycle as well
    substr_toks = decode_tokens(tokenizer, tokenizer(substring)["input_ids"])
    # we want to remove the start of sentence token if the tokenizer adds it
    if tokenizer.bos_token and substr_toks[0] == tokenizer.bos_token:
        substr_toks = substr_toks[1:]
    recoded_substr = "".join(substr_toks)
    toks = decode_tokens(tokenizer, token_array)
    whole_string = "".join(toks)
    char_locs = find_all_substring_indices(whole_string, recoded_substr)
    if len(char_locs) == 0:
        # sometimes adding a space in front causes different tokenization which works
        if substring[0] != " ":
            return find_token_range(tokenizer, token_array, " " + substring)
        raise ValueError(f"Could not find substring {recoded_substr} in {whole_string}")
    token_ranges: list[tuple[int, int]] = []
    for char_loc in char_locs:
        loc = 0
        tok_start, tok_end = None, None
        for i, t in enumerate(toks):
            loc += len(t)
            if tok_start is None and loc > char_loc:
                tok_start = i
            if tok_end is None and loc >= char_loc + len(recoded_substr):
                tok_end = i + 1
                break
        if tok_start is not None and tok_end is not None:
            token_ranges.append((tok_start, tok_end))
    if len(token_ranges) == 0:
        raise ValueError(f"Could not find substring {recoded_substr} in {toks}")
    return token_ranges[-1] if find_last_match else token_ranges[0]


def find_final_subject_token_index(
    tokenizer: Tokenizer, prompt: str, subject: str
) -> int:
    tokens = tokenizer.encode(prompt)
    _start, end = find_token_range(tokenizer, tokens, subject)
    return end - 1


def find_num_subject_tokens(tokenizer: Tokenizer, prompt: str, subject: str) -> int:
    tokens = tokenizer.encode(prompt)
    start, end = find_token_range(tokenizer, tokens, subject)
    return end - start


def predict_token(
    model: nn.Module, tokenizer: Tokenizer, prompts: list[str]
) -> tuple[list[str], torch.Tensor]:
    inp = make_inputs(tokenizer, prompts)
    preds, probs = predict_from_input(model, inp)
    results: list[str] = [tokenizer.decode(c) for c in preds]
    return (results, probs)


def predict_from_input(
    model: nn.Module,
    inp: dict[str, torch.Tensor],
    answer_id_overrides: list[tuple[int, int]] = [],
) -> tuple[torch.Tensor, torch.Tensor]:
    logits = predict_logits_from_input(model, inp)
    probs = torch.softmax(logits, dim=-1)
    prob, pred = torch.max(probs, dim=1)
    for i, j in answer_id_overrides:
        pred[i] = j
        prob[i] = probs[i, j]
    return pred, prob


def predict_logits_from_input(
    model: nn.Module,
    inp: dict[str, torch.Tensor],
) -> torch.Tensor:
    all_logits = model(**inp)["logits"]
    final_token_positions = find_final_attention_positions(inp["attention_mask"])
    batch_indices = torch.arange(all_logits.size(0))
    return all_logits[batch_indices, final_token_positions]


def predict_next_tokens_greedy(
    model: nn.Module,
    tokenizer: Tokenizer,
    prompts: list[str],
    num_tokens: int = 1,
    device: torch.device = DEFAULT_DEVICE,
) -> list[list[int]]:
    """
    Greedily predict the next N tokens for each prompt in the list.
    Should correctly handle right-padding.
    """
    next_prompts = [*prompts]  # copy to avoid modifying the original
    results: list[list[int]] = []
    for _i in range(num_tokens):
        # decoding and then re-encoding in a loop is wasteful, but it's the easiest way to handle
        # batches of different lengths, since model.generate() doesn't work with right-padding
        inputs = make_inputs(
            tokenizer,
            next_prompts,
            device=device,
        )
        pred_res = predict_from_input(model, inputs)
        for j, pred in enumerate(pred_res[0].detach().cpu()):
            if j >= len(results):
                results.append([])
            results[j].append(pred.item())
            next_prompts[j] += tokenizer.decode(pred)
    return results


def find_final_attention_positions(attention_mask: torch.Tensor) -> torch.Tensor:
    # convoluted, and generated by ChatGPT, but seems to work
    indices = torch.arange(attention_mask.size(1)).to(attention_mask.device)
    # use broadcasting to expand indices to the shape of attention_mask
    indices = indices[None, :].expand_as(attention_mask)
    # set indices where attention_mask is 0 to -1
    indices = torch.where(attention_mask == 1, indices, -1)
    # find the max indices
    max_indices = indices.max(dim=1).values
    return max_indices


def ensure_space_prefix(token_str: str) -> str:
    return " " + token_str.strip()


def answer_matches_expected(answer: str, expected_answers: Iterable[str]) -> bool:
    """
    Check if the given answer matches any of the expected answers. Handles case and whitespace.
    """
    processed_answer = process_answer(answer)
    return processed_answer in {process_answer(a) for a in expected_answers}


def process_answer(answer: str) -> str:
    """
    Process the given answer to make it easier to compare to other answers by removing case and trimming.
    """
    return answer.strip().lower()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['causal_tracer', 'causal_tracer.causal_tracing', 'causal_tracer.lib']

package_data = \
{'': ['*']}

install_requires = \
['matplotlib>=3.7.1,<4.0.0',
 'tqdm>=4.66.1,<5.0.0',
 'transformers>=4.28.1,<5.0.0']

setup_kwargs = {
    'name': 'causal-tracer',
    'version': '0.1.0',
    'description': '',
    'long_description': '# Causal Tracer\n\nCausal trace plots for transformer language models\n\n## About\n\nThis library is based heavily on causal tracing code from [ROME](https://rome.baulab.info/), and package and improves on their excellent work.\n',
    'author': 'David Chanin',
    'author_email': 'chanindav@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from collections import OrderedDict
from dataclasses import dataclass, field
from typing import Literal

import torch

LayerKind = Literal["hidden", "mlp", "attention"]


@dataclass
class HiddenFlow:
    scores: torch.Tensor
    low_score: float
    high_score: float
    input_ids: torch.Tensor
    input_tokens: list[str]
    subject_range: tuple[int, int]
    answer: str
    kind: LayerKind
    layer_outputs: OrderedDict[str, torch.Tensor] = field(repr=False)
